import math


def julian_day(
    year: int, month: int, day: int, hours: float = 0.0, minutes: float = 0.0
):
    if minutes != 0.0:
        hours = hours + (minutes / 60.0)

    Y = year if month > 2 else year - 1
    M = month if month > 2 else month + 12
    D = day + (hours / 24)

    A = math.floor(Y / 100)
    B = math.floor(2 - A + (A / 4))

    i0 = int(365.25 * (Y + 4716))
    i1 = int(30.6001 * (M + 1))

    return i0 + i1 + D + B - 1524.5


def julian_century(JD):
    # Equation from Astronomical Algorithms page 163
    return (JD - 2451545.0) / 36525



import calendar
from datetime import datetime, timedelta


def days_since_solstice(day_of_year: int, year: int, latitude: float) -> int:
    northern_offset = 10
    is_leap_year = calendar.isleap(year)

    southern_offset = 173 if is_leap_year else 172
    days_in_year = 366 if is_leap_year else 365

    if latitude >= 0:
        days_since_solstice = day_of_year + northern_offset
        if days_since_solstice >= days_in_year:
            days_since_solstice = days_since_solstice - days_in_year
    else:
        days_since_solstice = day_of_year - southern_offset
        if days_since_solstice < 0:
            days_since_solstice = days_since_solstice + days_in_year

    return days_since_solstice


def season_adjusted_morning_twilight(
    latitude: float, day_of_year: int, year: int, sunrise: datetime
):
    a = 75 + ((28.65 / 55.0) * abs(latitude))
    b = 75 + ((19.44 / 55.0) * abs(latitude))
    c = 75 + ((32.74 / 55.0) * abs(latitude))
    d = 75 + ((48.10 / 55.0) * abs(latitude))

    # final double adjustment;
    dyy = days_since_solstice(day_of_year, year, latitude)

    if dyy < 91:
        adjustment = a + (b - a) / 91.0 * dyy
    elif dyy < 137:
        adjustment = b + (c - b) / 46.0 * (dyy - 91)
    elif dyy < 183:
        adjustment = c + (d - c) / 46.0 * (dyy - 137)
    elif dyy < 229:
        adjustment = d + (c - d) / 46.0 * (dyy - 183)
    elif dyy < 275:
        adjustment = c + (b - c) / 46.0 * (dyy - 229)
    else:
        adjustment = b + (a - b) / 91.0 * (dyy - 275)

    return sunrise + timedelta(seconds=-int(round(adjustment * 60.0)))


def season_adjusted_evening_twilight(
    latitude: float, day: int, year: int, sunset: datetime
) -> datetime:
    a = 75 + ((25.60 / 55.0) * abs(latitude))
    b = 75 + ((2.050 / 55.0) * abs(latitude))
    c = 75 - ((9.210 / 55.0) * abs(latitude))
    d = 75 + ((6.140 / 55.0) * abs(latitude))

    dyy = days_since_solstice(day, year, latitude)
    if dyy < 91:
        adjustment = a + (b - a) / 91.0 * dyy
    elif dyy < 137:
        adjustment = b + (c - b) / 46.0 * (dyy - 91)
    elif dyy < 183:
        adjustment = c + (d - c) / 46.0 * (dyy - 137)
    elif dyy < 229:
        adjustment = d + (c - d) / 46.0 * (dyy - 183)
    elif dyy < 275:
        adjustment = c + (b - c) / 46.0 * (dyy - 229)
    else:
        adjustment = b + (a - b) / 91.0 * (dyy - 275)

    return sunset + timedelta(seconds=int(round(adjustment * 60.0)))



from dataclasses import dataclass


@dataclass
class DateComponents:
    year: int
    month: int
    day: int

    @classmethod
    def from_utc(cls, date):
        return cls(date.year, date.month, date.day)



import math
from adhanpy.astronomy.Astronomical import (
    approximate_transit,
    corrected_hour_angle,
    corrected_transit,
)
from adhanpy.astronomy.CalendricalHelper import julian_day
from adhanpy.data.ShadowLength import ShadowLength
from adhanpy.astronomy.SolarCoordinates import SolarCoordinates


class SolarTime:
    def __init__(self, date_components, coordinates):
        julian_date = julian_day(
            date_components.year, date_components.month, date_components.day
        )

        self.prev_solar = SolarCoordinates(julian_date - 1)
        self.solar = SolarCoordinates(julian_date)
        self.next_solar = SolarCoordinates(julian_date + 1)

        self.approximate_transit = approximate_transit(
            coordinates.longitude,
            self.solar.apparent_sidereal_time,
            self.solar.right_ascension,
        )
        solar_altitude = -50.0 / 60.0

        self.observer = coordinates
        self.transit = corrected_transit(
            self.approximate_transit,
            coordinates.longitude,
            self.solar.apparent_sidereal_time,
            self.solar.right_ascension,
            self.prev_solar.right_ascension,
            self.next_solar.right_ascension,
        )
        self.sunrise = corrected_hour_angle(
            self.approximate_transit,
            solar_altitude,
            coordinates,
            False,
            self.solar.apparent_sidereal_time,
            self.solar.right_ascension,
            self.prev_solar.right_ascension,
            self.next_solar.right_ascension,
            self.solar.declination,
            self.prev_solar.declination,
            self.next_solar.declination,
        )
        self.sunset = corrected_hour_angle(
            self.approximate_transit,
            solar_altitude,
            coordinates,
            True,
            self.solar.apparent_sidereal_time,
            self.solar.right_ascension,
            self.prev_solar.right_ascension,
            self.next_solar.right_ascension,
            self.solar.declination,
            self.prev_solar.declination,
            self.next_solar.declination,
        )

    def hour_angle(self, angle, after_transit):
        return corrected_hour_angle(
            self.approximate_transit,
            angle,
            self.observer,
            after_transit,
            self.solar.apparent_sidereal_time,
            self.solar.right_ascension,
            self.prev_solar.right_ascension,
            self.next_solar.right_ascension,
            self.solar.declination,
            self.prev_solar.declination,
            self.next_solar.declination,
        )

    def afternoon(self, shadow_length: ShadowLength):
        # TODO (from Swift version) source shadow angle calculation
        tangent = abs(self.observer.latitude - self.solar.declination)
        inverse = shadow_length.shadow_length + math.tan(math.radians(tangent))
        angle = math.degrees(math.atan(1.0 / inverse))

        return self.hour_angle(angle, True)



from enum import Enum


class CalculationMethod(Enum):

    NONE = 0

    MUSLIM_WORLD_LEAGUE = 1
    """
    Muslim World League
    Uses Fajr angle of 18 and an Isha angle of 17
    """

    EGYPTIAN = 2
    """
    Egyptian General Authority of Survey
    Uses Fajr angle of 19.5 and an Isha angle of 17.5
    """

    KARACHI = 3
    """
    University of Islamic Sciences, Karachi
    Uses Fajr angle of 18 and an Isha angle of 18
    """

    UMM_AL_QURA = 4
    """
    Umm al-Qura University, Makkah
    Uses a Fajr angle of 18.5 and an Isha angle of 90. Note: You should add a +30 minute custom
    adjustment of Isha during Ramadan.
    """

    DUBAI = 5
    """
    The Gulf Region
    Uses Fajr and Isha angles of 18.2 degrees.
    """

    MOON_SIGHTING_COMMITTEE = 6
    """
    Moonsighting Committee
    Uses a Fajr angle of 18 and an Isha angle of 18. Also uses seasonal adjustment values.
    """

    NORTH_AMERICA = 7
    """
    Referred to as the ISNA method
    This method is included for completeness, but is not recommended.
    Uses a Fajr angle of 15 and an Isha angle of 15.
    """

    KUWAIT = 8
    """
    Kuwait
    Uses a Fajr angle of 18 and an Isha angle of 17.5
    """

    QATAR = 9
    """
    Qatar
    Modified version of Umm al-Qura that uses a Fajr angle of 18.
    """

    SINGAPORE = 10
    """
    Singapore
    Uses a Fajr angle of 20 and an Isha angle of 18
    """

    UOIF = 11
    """
    UOIF
    Uses a Fajr angle of 12 and an Isha angle of 12
    """



from typing import Optional
from adhanpy.calculation.CalculationMethod import CalculationMethod
from adhanpy.calculation.MethodsParameters import methods_parameters
from adhanpy.calculation.Madhab import Madhab
from adhanpy.calculation.HighLatitudeRule import HighLatitudeRule
from adhanpy.calculation.PrayerAdjustments import PrayerAdjustments
from adhanpy.data.NightPortions import NightPortions


class CalculationParameters:
    def __init__(
        self,
        method: Optional[CalculationMethod] = None,
        adjustments: Optional[PrayerAdjustments] = None,
        method_adjustments: Optional[PrayerAdjustments] = None,
        isha_interval: int = 0,
        fajr_angle: float = 0.0,
        isha_angle: float = 0.0,
    ) -> None:
        # The madhab used to calculate Asr
        self.madhab = Madhab.SHAFI

        # Rules for placing bounds on Fajr and Isha for high latitude areas
        self.high_latitude_rule = HighLatitudeRule.MIDDLE_OF_THE_NIGHT

        # Minutes after Maghrib (if set, the time for Isha will be Maghrib plus isha_interval)
        self.isha_interval = isha_interval

        # fajr and isha angles
        self.fajr_angle = fajr_angle
        self.isha_angle = isha_angle

        # Used to optionally add or subtract a set amount of time from each prayer time
        self.adjustments = (
            adjustments if adjustments is not None else PrayerAdjustments()
        )

        # method is last assigned and has precedence and will overwrite other parameters
        self.method = (
            method if isinstance(method, CalculationMethod) else CalculationMethod.NONE
        )

        # Used for method adjustments
        self.method_adjustments = (
            method_adjustments
            if method_adjustments is not None
            else PrayerAdjustments()
        )

        self._set_parameters_using_method()

    def night_portions(self) -> NightPortions:
        if self.high_latitude_rule == HighLatitudeRule.MIDDLE_OF_THE_NIGHT:
            return NightPortions(1.0 / 2.0, 1.0 / 2.0)

        elif self.high_latitude_rule == HighLatitudeRule.SEVENTH_OF_THE_NIGHT:
            return NightPortions(1.0 / 7.0, 1.0 / 7.0)

        elif self.high_latitude_rule == HighLatitudeRule.TWILIGHT_ANGLE:
            return NightPortions(self.fajr_angle / 60.0, self.isha_angle / 60.0)

        raise ValueError("Invalid high latitude rule")

    def _set_parameters_using_method(self) -> None:
        method_parameters = methods_parameters[self.method]
        for key, value in method_parameters.items():
            setattr(self, key, value)



from enum import Enum


class HighLatitudeRule(Enum):

    MIDDLE_OF_THE_NIGHT = 0
    """
    Fajr will never be earlier than the middle of the night, and Isha will never be later than
    the middle of the night.
    """

    SEVENTH_OF_THE_NIGHT = 1
    """
    Fajr will never be earlier than the beginning of the last seventh of the night, and Isha will
    never be later than the end of hte first seventh of the night.
    """

    TWILIGHT_ANGLE = 2
    """
    Similar to {@link HighLatitudeRule#SEVENTH_OF_THE_NIGHT}, but instead of 1/7th, the faction
    of the night used is fajrAngle / 60 and ishaAngle/60.
    """



from __future__ import annotations
import math
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Optional
from adhanpy.util.DateComponents import DateComponents


@dataclass
class TimeComponents:
    hours: int
    minutes: int
    seconds: int

    @classmethod
    def from_float(cls, value: float) -> Optional[TimeComponents]:
        if math.isinf(value) or math.isnan(value):
            return None

        minutes, seconds = divmod(value * 60 * 60, 60)
        hours, minutes = divmod(minutes, 60)
        return cls(int(hours), int(minutes), int(seconds))

    def date_components(self, date_components: DateComponents) -> datetime:
        date_time = datetime(
            date_components.year,
            date_components.month,
            date_components.day,
            0,
            self.minutes,
            self.seconds,
            tzinfo=timezone.utc,
        )
        date_time = date_time + timedelta(hours=self.hours)

        return date_time



class PrayerAdjustments:

    fajr: int
    # Fajr offset in minutes

    sunrise: int
    # Sunrise offset in minutes

    dhuhr: int
    # Dhuhr offset in minutes

    asr: int
    # Asr offset in minutes

    maghrib: int
    # Maghrib offset in minutes

    isha: int
    # Isha offset in minutes

    def __init__(
        self,
        fajr: int = 0,
        sunrise: int = 0,
        dhuhr: int = 0,
        asr: int = 0,
        maghrib: int = 0,
        isha: int = 0,
    ):
        """
        Gets a PrayerAdjustments object to offset prayer times (defaulting to 0)
        param fajr offset from fajr in minutes
        param sunrise offset from sunrise in minutes
        param dhuhr offset from dhuhr in minutes
        param asr offset from asr in minutes
        param maghrib offset from maghrib in minutes
        param isha offset from isha in minutes
        """
        self.fajr = fajr
        self.sunrise = sunrise
        self.dhuhr = dhuhr
        self.asr = asr
        self.maghrib = maghrib
        self.isha = isha



from datetime import datetime, timedelta, timezone
from zoneinfo import ZoneInfo
from typing import Tuple
from adhanpy.calculation.CalculationMethod import CalculationMethod
from adhanpy.calculation.CalculationParameters import CalculationParameters
from adhanpy.calculation.Twilight import (
    season_adjusted_evening_twilight,
    season_adjusted_morning_twilight,
)
from adhanpy.data.Prayer import Prayer
from adhanpy.astronomy.SolarTime import SolarTime
from adhanpy.data.Coordinates import Coordinates
from adhanpy.util.TimeComponents import TimeComponents
from adhanpy.util.DateComponents import DateComponents
from adhanpy.util.CalendarUtil import rounded_minute


class PrayerTimes:
    def __init__(
        self,
        coordinates: Tuple[float, float],
        date: datetime,
        calculation_method: CalculationMethod = None,
        calculation_parameters: CalculationParameters = None,
        time_zone: ZoneInfo = None,
    ):
        """
        Arguments:
            coordinates: (latitude, longitude)
            date: DateComponents
            calculation_parameters: CalculationParameters
            time_zone: example ZoneInfo("Europe/London")
        Returns:
            PrayerTimes object with UTC datetimes for fajr, sunrise, dhuhr, asr, maghrib and isha
        """

        if (calculation_parameters and calculation_method) or not (
            calculation_parameters or calculation_method
        ):
            raise ValueError(
                "Only one of calculation_method or calculation_parameters must be passed."
            )

        self.calculation_parameters = calculation_parameters

        if self.calculation_parameters is None:
            self.calculation_parameters = CalculationParameters(
                method=calculation_method
            )

        latitude, longitude = coordinates
        self.coordinates = Coordinates(latitude, longitude)
        self._date_components = DateComponents.from_utc(date)
        self.time_zone = time_zone

        self._prayer_date = datetime(
            self._date_components.year,
            self._date_components.month,
            self._date_components.day,
            tzinfo=timezone.utc,
        )

        self._day_of_year = self._prayer_date.timetuple().tm_yday

        tomorrow_date = self._prayer_date + timedelta(days=1)
        tomorrow_date_components = DateComponents.from_utc(tomorrow_date)

        self._solar_time = SolarTime(self._date_components, self.coordinates)

        time_components = TimeComponents.from_float(self._solar_time.transit)
        transit = (
            None
            if time_components is None
            else time_components.date_components(self._date_components)
        )

        time_components = TimeComponents.from_float(self._solar_time.sunrise)
        self._sunrise_components = (
            None
            if time_components is None
            else time_components.date_components(self._date_components)
        )

        time_components = TimeComponents.from_float(self._solar_time.sunset)
        self._sunset_components = (
            None
            if time_components is None
            else time_components.date_components(self._date_components)
        )

        tomorrow_solar_time = SolarTime(tomorrow_date_components, self.coordinates)
        tomorrow_sunrise_components = TimeComponents.from_float(
            tomorrow_solar_time.sunrise
        )

        if (
            transit is None
            or self._sunrise_components is None
            or self._sunset_components is None
            or tomorrow_sunrise_components is None
        ):
            raise RuntimeError

        # get night length
        tomorrow_sunrise = tomorrow_sunrise_components.date_components(
            tomorrow_date_components
        )
        self.night_length = (
            tomorrow_sunrise.timestamp() * 1000
            - self._sunset_components.timestamp() * 1000
        )
        self.night_portions = self.calculation_parameters.night_portions()

        # Assign final times to properties with all offsets
        self._set_fajr()
        self._set_sunrise()
        self._set_dhuhr(transit)
        self._set_asr()
        self._set_maghrib()
        self._set_isha(self._sunset_components)

        self._adjust_prayers_time_zone()

    def _set_fajr(self):
        temp_fajr = None
        if time_components := TimeComponents.from_float(
            self._solar_time.hour_angle(-self.calculation_parameters.fajr_angle, False)
        ):
            temp_fajr = time_components.date_components(self._date_components)

        if (
            self.calculation_parameters.method
            == CalculationMethod.MOON_SIGHTING_COMMITTEE
        ):
            if self.coordinates.latitude >= 55:
                temp_fajr = self._sunrise_components + timedelta(
                    seconds=-1 * int(self.night_length / 7000)
                )

            safe_fajr = season_adjusted_morning_twilight(
                self.coordinates.latitude,
                self._day_of_year,
                self._prayer_date.year,
                self._sunrise_components,
            )
        else:
            portion = self.night_portions.fajr
            night_fraction = int(portion * self.night_length / 1000)
            safe_fajr = self._sunrise_components + timedelta(
                seconds=-1 * night_fraction
            )

        if temp_fajr is None or temp_fajr < safe_fajr:
            temp_fajr = safe_fajr

        self.fajr = self._rounded_minute(
            self.calculation_parameters.adjustments,
            self.calculation_parameters.method_adjustments,
            "fajr",
            temp_fajr,
        )

    def _set_sunrise(self):
        self.sunrise = self._rounded_minute(
            self.calculation_parameters.adjustments,
            self.calculation_parameters.method_adjustments,
            "sunrise",
            self._sunrise_components,
        )

    def _set_dhuhr(self, time):
        self.dhuhr = self._rounded_minute(
            self.calculation_parameters.adjustments,
            self.calculation_parameters.method_adjustments,
            "dhuhr",
            time,
        )

    def _set_asr(self):
        if time_components := TimeComponents.from_float(
            self._solar_time.afternoon(
                self.calculation_parameters.madhab.get_shadow_length()
            )
        ):
            if temp_asr := time_components.date_components(self._date_components):
                self.asr = self._rounded_minute(
                    self.calculation_parameters.adjustments,
                    self.calculation_parameters.method_adjustments,
                    "asr",
                    temp_asr,
                )
        try:
            self.asr.hour
        except:
            raise RuntimeError

    def _set_maghrib(self):
        self.maghrib = self._rounded_minute(
            self.calculation_parameters.adjustments,
            self.calculation_parameters.method_adjustments,
            "maghrib",
            self._sunset_components,
        )

    def _set_isha(self, sunset):
        # Isha calculation with check against safe value
        temp_isha = None
        try:
            if self.calculation_parameters.isha_interval < 1:
                raise ValueError("Isha interval is either not defined or less than 1.")

            temp_isha = sunset + timedelta(
                seconds=self.calculation_parameters.isha_interval * 60
            )
        except:
            timeComponents = TimeComponents.from_float(
                self._solar_time.hour_angle(
                    -self.calculation_parameters.isha_angle, True
                )
            )

            if timeComponents is not None:
                temp_isha = timeComponents.date_components(self._date_components)

            if (
                self.calculation_parameters.method
                == CalculationMethod.MOON_SIGHTING_COMMITTEE
                and self.coordinates.latitude >= 55
            ):
                night_fraction = int(self.night_length / 7000)
                temp_isha = self._sunset_components + timedelta(seconds=night_fraction)

            if (
                self.calculation_parameters.method
                == CalculationMethod.MOON_SIGHTING_COMMITTEE
            ):
                safe_isha = season_adjusted_evening_twilight(
                    self.coordinates.latitude,
                    self._day_of_year,
                    self._date_components.year,
                    self._sunset_components,
                )
            else:
                portion = self.night_portions.isha
                night_fraction = int(portion * self.night_length / 1000)

                safe_isha = self._sunset_components + timedelta(
                    seconds=int(night_fraction)
                )

            if temp_isha is None or temp_isha > safe_isha:
                temp_isha = safe_isha

        self.isha = self._rounded_minute(
            self.calculation_parameters.adjustments,
            self.calculation_parameters.method_adjustments,
            "isha",
            temp_isha,
        )

    def _rounded_minute(
        self, adjustments, method_adjustments, prayer_name, temp_prayer
    ):
        prayer_adjustments = getattr(adjustments, prayer_name)
        method_prayer_adjustments = getattr(method_adjustments, prayer_name)
        return rounded_minute(
            (temp_prayer + timedelta(minutes=prayer_adjustments))
            + timedelta(minutes=method_prayer_adjustments)
        )

    def _adjust_prayers_time_zone(self):
        if self.time_zone is not None:
            self.fajr = self.fajr.astimezone(self.time_zone)
            self.sunrise = self.sunrise.astimezone(self.time_zone)
            self.dhuhr = self.dhuhr.astimezone(self.time_zone)
            self.asr = self.asr.astimezone(self.time_zone)
            self.maghrib = self.maghrib.astimezone(self.time_zone)
            self.isha = self.isha.astimezone(self.time_zone)



class ShadowLength:
    SINGLE = 1.0
    DOUBLE = 2.0

    def __init__(self, shadow_length):
        self._shadow_length = shadow_length

    @property
    def shadow_length(self):
        return self._shadow_length



from datetime import datetime


def rounded_minute(when: datetime) -> datetime:
    """
    Round the seconds of a datetime object to 0 or 1 minute
    and add it to the datetime's minutes when possible or
    drop the seconds
    when: datetime object
    return: datetime object with seconds rounded to the nearest minute
    """
    minute = when.minute
    second = when.second

    try:
        rounded = when.replace(minute=int(minute + round(second / 60)), second=0)
    except ValueError:
        rounded = when.replace(second=0)

    return rounded



from enum import Enum
from adhanpy.data.ShadowLength import ShadowLength


class Madhab(Enum):

    SHAFI = 0
    # Shafi Madhab

    HANAFI = 1
    # Hanafi Madhab

    def get_shadow_length(self) -> ShadowLength:
        if self == Madhab.SHAFI:
            return ShadowLength(ShadowLength.SINGLE)
        elif self == Madhab.HANAFI:
            return ShadowLength(ShadowLength.DOUBLE)



import math


def normalize_with_bound(value, max):
    return value - (max * (math.floor(value / max)))


def unwind_angle(value):
    return normalize_with_bound(value, 360)


def closest_angle(angle):
    if angle >= -180 and angle <= 180:
        return angle

    return angle - (360 * round(angle / 360))



from adhanpy.data.Coordinates import Coordinates
from adhanpy.util.FloatUtil import closest_angle, unwind_angle, normalize_with_bound
import math


def mean_solar_longitude(T: float) -> float:
    # Equation from Astronomical Algorithms page 163
    term1 = 280.4664567
    term2 = 36000.76983 * T
    term3 = 0.0003032 * (T**2)
    L0 = term1 + term2 + term3
    return unwind_angle(L0)


def mean_lunar_longitude(T: float) -> float:
    # Equation from Astronomical Algorithms page 144
    term1 = 218.3165
    term2 = 481267.8813 * T
    Lp = term1 + term2
    return unwind_angle(Lp)


def ascending_lunar_node_longitude(T: float) -> float:
    # Equation from Astronomical Algorithms page 144
    term1 = 125.04452
    term2 = 1934.136261 * T
    term3 = 0.0020708 * (T**2)
    term4 = (T**3) / 450000
    Ω = term1 - term2 + term3 + term4
    return unwind_angle(Ω)


def apparent_solar_longitude(T: float, L0: float) -> float:
    longitude = L0 + solar_equation_of_the_center(T, mean_solar_anomaly(T))
    Ω = 125.04 - (1934.136 * T)
    λ = longitude - 0.00569 - (0.00478 * math.sin(math.radians(Ω)))
    return unwind_angle(λ)


def solar_equation_of_the_center(T: float, M: float) -> float:
    # Equation from Astronomical Algorithms page 164
    Mrad = math.radians(M)
    term1 = (1.914602 - (0.004817 * T) - (0.000014 * (T**2))) * math.sin(Mrad)
    term2 = (0.019993 - (0.000101 * T)) * math.sin(2 * Mrad)
    term3 = 0.000289 * math.sin(3 * Mrad)
    return term1 + term2 + term3


def mean_solar_anomaly(T: float) -> float:
    # Equation from Astronomical Algorithms page 163
    term1 = 357.52911
    term2 = 35999.05029 * T
    term3 = 0.0001537 * (T**2)
    M = term1 + term2 - term3
    return unwind_angle(M)


def mean_sidereal_time(T: float) -> float:
    # Equation from Astronomical Algorithms page 165
    JD = (T * 36525) + 2451545.0
    term1 = 280.46061837
    term2 = 360.98564736629 * (JD - 2451545)
    term3 = 0.000387933 * (T**2)
    term4 = (T**3) / 38710000
    θ = term1 + term2 + term3 - term4
    return unwind_angle(θ)


def nutation_in_longitude(L0: float, Lp: float, Ω: float) -> float:
    # Equation from Astronomical Algorithms page 144
    term1 = (-17.2 / 3600) * math.sin(math.radians(Ω))
    term2 = (1.32 / 3600) * math.sin(2 * math.radians(L0))
    term3 = (0.23 / 3600) * math.sin(2 * math.radians(Lp))
    term4 = (0.21 / 3600) * math.sin(2 * math.radians(Ω))
    return term1 - term2 - term3 + term4


def nutation_in_obliquity(L0: float, Lp: float, Ω: float) -> float:
    # Equation from Astronomical Algorithms page 144
    term1 = (9.2 / 3600) * math.cos(math.radians(Ω))
    term2 = (0.57 / 3600) * math.cos(2 * math.radians(L0))
    term3 = (0.10 / 3600) * math.cos(2 * math.radians(Lp))
    term4 = (0.09 / 3600) * math.cos(2 * math.radians(Ω))
    return term1 + term2 + term3 - term4


def mean_obliquity_of_the_ecliptic(T: float) -> float:
    # Equation from Astronomical Algorithms page 147
    term1 = 23.439291
    term2 = 0.013004167 * T
    term3 = 0.0000001639 * (T**2)
    term4 = 0.0000005036 * (T**3)
    return term1 - term2 - term3 + term4


def apparent_obliquity_of_the_ecliptic(T: float, ε0: float) -> float:
    # Equation from Astronomical Algorithms page 165
    O = 125.04 - (1934.136 * T)
    return ε0 + (0.00256 * math.cos(math.radians(O)))


def altitude_of_celestial_body(φ: float, δ: float, H: float) -> float:
    # Equation from Astronomical Algorithms page 93
    term1 = math.sin(math.radians(φ)) * math.sin(math.radians(δ))
    term2 = (
        math.cos(math.radians(φ))
        * math.cos(math.radians(δ))
        * math.cos(math.radians(H))
    )
    return math.degrees(math.asin(term1 + term2))


def approximate_transit(L: float, Θ0: float, α2: float) -> float:
    # Equation from page Astronomical Algorithms 102
    Lw = L * -1
    return normalize_with_bound((α2 + Lw - Θ0) / 360, 1)


def corrected_transit(
    m0: float, L: float, Θ0: float, α2: float, α1: float, α3: float
) -> float:
    # Equation from page Astronomical Algorithms 102
    Lw = L * -1
    θ = unwind_angle(Θ0 + (360.985647 * m0))
    α = unwind_angle(interpolate_angles(α2, α1, α3, m0))
    H = closest_angle(θ - Lw - α)
    Δm = H / -360
    return (m0 + Δm) * 24


def interpolate(y2: float, y1: float, y3: float, n: float) -> float:
    # Equation from Astronomical Algorithms page 24
    a = y2 - y1
    b = y3 - y2
    c = b - a
    return y2 + ((n / 2) * (a + b + (n * c)))


def interpolate_angles(y2: float, y1: float, y3: float, n: float) -> float:
    # Equation from Astronomical Algorithms page 24
    a = unwind_angle(y2 - y1)
    b = unwind_angle(y3 - y2)
    c = b - a
    return y2 + ((n / 2) * (a + b + (n * c)))


def corrected_hour_angle(
    m0: float,
    h0: float,
    coordinates: Coordinates,
    afterTransit: bool,
    Θ0: float,
    α2: float,
    α1: float,
    α3: float,
    δ2: float,
    δ1: float,
    δ3: float,
) -> float:
    # Equation from page Astronomical Algorithms 102
    Lw = coordinates.longitude * -1
    term1 = math.sin(math.radians(h0)) - (
        math.sin(math.radians(coordinates.latitude)) * math.sin(math.radians(δ2))
    )
    term2 = math.cos(math.radians(coordinates.latitude)) * math.cos(math.radians(δ2))
    try:
        H0 = math.degrees(math.acos(term1 / term2))
        m = m0 + (H0 / 360) if afterTransit else m0 - (H0 / 360)
        θ = unwind_angle(Θ0 + (360.985647 * m))
        α = unwind_angle(interpolate_angles(α2, α1, α3, m))
        δ = interpolate(δ2, δ1, δ3, m)
        H = θ - Lw - α
        h = altitude_of_celestial_body(coordinates.latitude, δ, H)
        term3 = h - h0
        term4 = (
            360
            * math.cos(math.radians(δ))
            * math.cos(math.radians(coordinates.latitude))
            * math.sin(math.radians(H))
        )
        Δm = term3 / term4
    except:
        return math.nan

    return (m + Δm) * 24



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="adhanpy",
    version="1.0.5",
    author="alphahm",
    url="https://github.com/alphahm/adhanpy",
    description="An offline library calculating prayer times",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    project_urls={
        'Documentation': "https://github.com/alphahm/adhanpy/blob/master/README.md",
        'Changelog': 'https://github.com/alphahm/adhanpy/blob/master/CHANGES.md',
    },
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import math
from adhanpy.astronomy.CalendricalHelper import julian_century
from adhanpy.astronomy.Astronomical import (
    apparent_obliquity_of_the_ecliptic,
    apparent_solar_longitude,
    mean_obliquity_of_the_ecliptic,
    mean_sidereal_time,
    mean_solar_longitude,
    mean_lunar_longitude,
    ascending_lunar_node_longitude,
    nutation_in_longitude,
    nutation_in_obliquity,
)
from adhanpy.util.FloatUtil import unwind_angle


class SolarCoordinates:
    def __init__(self, julian_day) -> None:
        T = julian_century(julian_day)
        L0 = mean_solar_longitude(T)
        Lp = mean_lunar_longitude(T)
        Ω = ascending_lunar_node_longitude(T)
        λ = math.radians(apparent_solar_longitude(T, L0))
        θ0 = mean_sidereal_time(T)
        ΔΨ = nutation_in_longitude(L0, Lp, Ω)
        Δε = nutation_in_obliquity(L0, Lp, Ω)
        ε0 = mean_obliquity_of_the_ecliptic(T)
        εapp = math.radians(apparent_obliquity_of_the_ecliptic(T, ε0))

        # Equation from Astronomical Algorithms page 165
        self.declination = math.degrees(math.asin(math.sin(εapp) * math.sin(λ)))

        # Equation from Astronomical Algorithms page 165
        self.right_ascension = unwind_angle(
            math.degrees(math.atan2(math.cos(εapp) * math.sin(λ), math.cos(λ)))
        )

        # Equation from Astronomical Algorithms page 88
        self.apparent_sidereal_time = θ0 + (
            ((ΔΨ * 3600) * math.cos(math.radians(ε0 + Δε))) / 3600
        )



from adhanpy.calculation.CalculationMethod import CalculationMethod
from adhanpy.calculation.PrayerAdjustments import PrayerAdjustments


methods_parameters = {
    CalculationMethod.NONE: {},
    CalculationMethod.MUSLIM_WORLD_LEAGUE: {
        "fajr_angle": 18.0,
        "isha_angle": 17.0,
        "method_adjustments": PrayerAdjustments(dhuhr=1),
    },
    CalculationMethod.EGYPTIAN: {
        "fajr_angle": 19.5,
        "isha_angle": 17.5,
        "method_adjustments": PrayerAdjustments(dhuhr=1),
    },
    CalculationMethod.KARACHI: {
        "fajr_angle": 18.0,
        "isha_angle": 18.0,
        "method_adjustments": PrayerAdjustments(dhuhr=1),
    },
    CalculationMethod.UMM_AL_QURA: {"fajr_angle": 18.5, "isha_interval": 90},
    CalculationMethod.DUBAI: {
        "fajr_angle": 18.2,
        "isha_angle": 18.2,
        "method_adjustments": PrayerAdjustments(sunrise=-3, dhuhr=3, asr=3, maghrib=3),
    },
    CalculationMethod.MOON_SIGHTING_COMMITTEE: {
        "fajr_angle": 18.0,
        "isha_angle": 18.0,
        "method_adjustments": PrayerAdjustments(dhuhr=5, maghrib=3),
    },
    CalculationMethod.NORTH_AMERICA: {
        "fajr_angle": 15.0,
        "isha_angle": 15.0,
        "method_adjustments": PrayerAdjustments(dhuhr=1),
    },
    CalculationMethod.KUWAIT: {"fajr_angle": 18.0, "isha_angle": 17.5},
    CalculationMethod.QATAR: {"fajr_angle": 18.0, "isha_interval": 90},
    CalculationMethod.SINGAPORE: {
        "fajr_angle": 20.0,
        "isha_angle": 18.0,
        "method_adjustments": PrayerAdjustments(dhuhr=1),
    },
    CalculationMethod.UOIF: {"fajr_angle": 12.0, "isha_angle": 12.0},
}
